<?php

namespace DigitalPoint\Cloudflare\Api;
abstract class Advanced extends CloudflareAbstract
{
	protected $r2BaseUrl = 'https://%s.r2.cloudflarestorage.com/%s/';

	protected function r2BaseUrlFull($bucketName)
	{
		$accountId = $this->option('cfAccountId');
		if (!$accountId)
		{
			$message = 'Cloudflare account ID missing. Double check that all required permissions are set for your API token.';
			$this->logError($message);
			$this->printableException($message);
		}

		return sprintf($this->r2BaseUrl, $accountId, $bucketName);
	}
	protected function getActionsByType($actionType = 'classA')
	{
		$actionTypeGroups = [
			'classA' => [
				"ListBuckets",
				"PutBucket",
				"ListObjects",
				"PutObject",
				"CopyObject",
				"CompleteMultipartUpload",
				"CreateMultipartUpload",
				"UploadPart",
				"UploadPartCopy",
				"PutBucketEncryption",
				"ListMultipartUploads"
			],
			'classB' => [
				"HeadBucket",
				"HeadObject",
				"GetObject",
				"ReportUsageSummary",
				"GetBucketEncryption",
				"GetBucketLocation"
			]
		];

		if (!empty($actionTypeGroups[$actionType]))
		{
			$actionTypeOperator = 'actionType_in';
			$actionTypes = $actionTypeGroups[$actionType];
		}
		else
		{
			$actionTypeOperator = 'actionType_notin';
			$actionTypes = [];

			foreach ($actionTypeGroups as $group)
			{
				$actionTypes = array_merge($actionTypes, $group);
			}
		}

		return [
			$actionTypes,
			$actionTypeOperator
		];
	}
	public function getGraphQLR2OperationsLog($accountId, $bucketName = null, $actionType = 'classA')
	{
		$dateFrom = new \DateTime();
		$dateFrom->setTimestamp($this->time());
		$dateFrom->sub(new \DateInterval('P14D'));

		list($actionTypes, $actionTypeOperator) = $this->getActionsByType($actionType);

		$query = 'query getR2OperationsQuery($accountTag: string, $filter: AccountR2OperationsAdaptiveGroupsFilter_InputObject) {
  viewer {
    accounts(filter: {accountTag: $accountTag}) {
      r2OperationsAdaptiveGroups(limit: 500, orderBy: $orderBy, filter: $filter) {
        dimensions {
          actionStatus
          actionType
          bucketName
          datetime
          objectName
          responseStatusCode
          __typename
        }
        __typename
      }
      __typename
    }
    __typename
  }
}';

		$variables = [
			'accountTag' => $accountId,
			'orderBy' => ['datetime_DESC'],
			'filter' => [
				'AND' => [
					[
						//	'actionStatus' => 'success',
						'date_geq' => $dateFrom->format('Y-m-d'),
						$actionTypeOperator => $actionTypes,
					],
				],
			],
		];

		if ($bucketName)
		{
			$variables['filter']['AND'][0]['bucketName'] = $bucketName;
		}

		return $this->queryGraphQL(
			'getR2OperationsQuery',
			$query,
			$variables
		);
	}

	public function getGraphQLR2Storage($accountId, $bucketName = null)
	{
		$date = new \DateTime();
		$date->setTimestamp($this->time());
		$date->sub(new \DateInterval('P7D'));

		$query = 'query getR2StorageQuery($accountTag: string, $filter: AccountR2StorageAdaptiveGroupsFilter_InputObject) {
  viewer {
    accounts(filter: {accountTag: $accountTag}) {
      r2StorageAdaptiveGroups(limit: 10000, orderBy: $orderBy, filter: $filter) {
      	dimensions {
      	  bucketName
      	  datetimeMinute
      	  __typename
      	}
        max {
          metadataSize
          objectCount
          payloadSize
          __typename
        }
        __typename
      }
      __typename
    }
    __typename
  }
}';

		$variables = [
			'accountTag' => $accountId,
			'orderBy' => ['datetimeMinute_DESC'],
			'filter' => [
				'AND' => [
					[
						'date_geq' => $date->format('Y-m-d'),
					],
				],
			],
		];

		if ($bucketName)
		{
			$variables['filter']['AND'][0]['bucketName'] = $bucketName;
		}

		return $this->queryGraphQL(
			'getR2StorageQuery',
			$query,
			$variables
		);
	}

	public function getGraphQLR2Operations($accountId, $bucketName = null, $actionType = 'classA', $exactDate = null)
	{
		list($actionTypes, $actionTypeOperator) = $this->getActionsByType($actionType);

		$query = 'query getR2OperationsQuery($accountTag: string, $filter: AccountR2OperationsAdaptiveGroupsFilter_InputObject) {
  viewer {
    accounts(filter: {accountTag: $accountTag}) {
      r2OperationsAdaptiveGroups(limit: 10000, filter: $filter) {
        sum {
          requests
          responseObjectSize
          __typename
        }
        __typename
      }
      __typename
    }
    __typename
  }
}';

		$variables = [
			'accountTag' => $accountId,
			'filter' => [
				'AND' => [
					[
						//		'actionStatus' => 'success',
						$actionTypeOperator => $actionTypes,
						//				'date_geq' => $currentDate->format('Y-m-01'),
						//				'date_leq' => $currentDate->format('Y-m-d'),
					],
				],
			],
		];

		if ($exactDate)
		{
			$date = new \DateTime($exactDate);

			$variables['filter']['AND'][0]['date'] = $date->format('Y-m-d');
		}
		else
		{
			$currentDate = new \DateTime();
			$currentDate->setTimestamp($this->time());

			$variables['filter']['AND'][0]['date_geq'] = $currentDate->format('Y-m-01');
			$variables['filter']['AND'][0]['date_leq'] = $currentDate->format('Y-m-d');
		}

		if ($bucketName)
		{
			if (is_array($bucketName))
			{
				$variables['filter']['AND'][0]['bucketName_in'] = $bucketName;
			}
			else
			{
				$variables['filter']['AND'][0]['bucketName'] = $bucketName;
			}
		}

		return $this->queryGraphQL(
			'getR2OperationsQuery',
			$query,
			$variables
		);
	}

	public function getZoneSubscription($zoneId, $returnPromise = true)
	{
		$method = $returnPromise ? 'getAsync' : 'GET';

		$params = [];
		return $this->makeRequest($method, sprintf('zones/%s/subscription', $zoneId), ['query' => $params]);
	}

	public function createTurnstileWidget($accountId, $name, $domains, $mode = 'managed')
	{
		if (is_string($domains))
		{
			$domains = [$domains];
		}
		$params = [
			'name' => $name,
			'mode' => $mode,
			'domains' => $domains,
		];
		return $this->makeRequest('POST', sprintf('accounts/%s/challenges/widgets', $accountId), ['json' => $params]);
	}

	public function getTurnstileAnalytics($accountId, $siteKey, $date = null)
	{
		if (!$date)
		{
			$date = new \DateTime();
			$date->setTimestamp($this->time());
		}
		else
		{
			$date = new \DateTime($date);
		}

		$query = 'query widgetAnalyticsQuery($accountTag: string, $filter: AccountTurnstileAdaptiveGroupsFilter_InputObject) {
			viewer {
				accounts(filter: {accountTag: $accountTag}) {
					solveEvents: turnstileAdaptiveGroups(filter: $filter, limit: 1000) {
						count
						dimensions {
							eventType
							timestamp: datetimeHour
							__typename
						}
						__typename
					}
					__typename
				}
				__typename
			}
		}';
		$variables = [
			'accountTag' => $accountId,
			'filter' => [
				'AND' => [
					[
						'siteKey' => $siteKey,
						'date' => $date->format('Y-m-d'),
					],
				],
			],
		];

		return $this->queryGraphQL(
			'widgetAnalyticsQuery',
			$query,
			$variables
		);
	}

	public function createFirewallRule($zoneId, $rulesetId, $expression, $description = '', $ref = '', $enabled = true, $action = 'block')
	{
		$params = [
			'action' => $action,
			'expression' => $expression,
			'description' => $description,
			'enabled' => $enabled,
			'ref' => $ref
		];

		if ($rulesetId)
		{
			$return = $this->makeRequest('POST', sprintf('zones/%s/rulesets/%s/rules', $zoneId, $rulesetId), ['json' => $params]);
		}
		else
		{
			$params = ['rules' => [
				$params
			]];

			$return = $this->makeRequest('PUT', sprintf('zones/%s/rulesets/phases/http_request_firewall_custom/entrypoint', $zoneId), ['json' => $params]);
		}

		return $return;
	}

	public function updateFirewallRule($zoneId, $rulesetId, $identifier, $action, $expression, $description, $enabled, $lastUpdated, $version, $ref = '')
	{
		$params = [];

		if ($ref)
		{
			$params['ref'] = $ref;
		}

		$params['id'] = $identifier;
		$params['action'] = $action;
		$params['expression'] = $expression;
		$params['description'] = $description;
		$params['enabled'] = $enabled;
		$params['last_updated'] = $lastUpdated;
		$params['version'] = $version;

		try {
			$return = $this->makeRequest('PATCH', sprintf('zones/%s/rulesets/%s/rules/%s', $zoneId, $rulesetId, $identifier), ['json' => $params]);
		}
		catch(\Exception $e)
		{
			$return = null;
		}

		return $return;
	}

	public function deleteFirewallRule($zoneId, $rulesetId, $identifier)
	{
		return $this->makeRequest('DELETE', sprintf('zones/%s/rulesets/%s/rules/%s', $zoneId, $rulesetId, $identifier), ['query' => []]);
	}

	public function createFirewallUserAgentRule($zoneId, array $configuration, $mode = 'block', $description = '')
	{
		$params = [
			'mode' => $mode,
			'configuration' => $configuration,
			'description' => $description
		];

		try {
			$return = $this->makeRequest('POST', sprintf('zones/%s/firewall/ua_rules', $zoneId), ['json' => $params]);
		}
		catch(\Exception $e)
		{
			$return = null;
		}

		return $return;
	}

	public function updateFirewallUserAgentRule($zoneId, $identifier, $mode, array $configuration, $description, $paused)
	{
		$params = [];

		$params['id'] = $identifier;
		$params['mode'] = $mode;
		$params['configuration'] = $configuration;
		$params['description'] = $description;
		$params['paused'] = $paused;

		try {
			$return = $this->makeRequest('PUT', sprintf('zones/%s/firewall/ua_rules/%s', $zoneId, $identifier), ['json' => $params]);
		}
		catch(\Exception $e)
		{
			$return = null;
		}

		return $return;
	}

	public function deleteFirewallUserAgentRule($zoneId, $identifier)
	{
		return $this->makeRequest('DELETE', sprintf('zones/%s/firewall/ua_rules/%s', $zoneId, $identifier), ['json' => []]);
	}


	public function createPageRule($zoneId, array $targets, array $actions, $priority = 1, $status = 'active')
	{
		$params = [
			'targets' => $targets,
			'actions' => $actions,
			'priority' => $priority,
			'status' => $status
		];

		try {
			$return = $this->makeRequest('POST', sprintf('zones/%s/pagerules', $zoneId), ['json' => $params]);
		}
		catch(\Exception $e)
		{
			$return = null;
		}

		return $return;
	}


	public function updatePageRule($zoneId, $identifier, $status, array $targets, $actions, $priority)
	{
		$params = [];

		$params['id'] = $identifier;
		$params['actions'] = $actions;
		$params['targets'] = $targets;
		$params['status'] = $status;
		$params['priority'] = intval($priority);

		try {
			$return = $this->makeRequest('PUT', sprintf('zones/%s/pagerules/%s', $zoneId, $identifier), ['json' => $params]);
		}
		catch(\Exception $e)
		{
			$return = null;
		}

		return $return;
	}


	public function deletePageRule($zoneId, $identifier)
	{
		return $this->makeRequest('DELETE', sprintf('zones/%s/pagerules/%s', $zoneId, $identifier), ['json' => []]);
	}

	public function updateCacheRule($zoneId, $rulesetId, $identifier, $action, $actionParameters, $description, $enabled, $expression)
	{
		$params = [];

		$params['id'] = $identifier;
		$params['action'] = $action;
		$params['action_parameters'] = $actionParameters;
		$params['description'] = $description;
		$params['enabled'] = (bool)$enabled;
		$params['expression'] = $expression;

		try {
			$return = $this->makeRequest('PATCH', sprintf('zones/%s/rulesets/%s/rules/%s', $zoneId, $rulesetId, $identifier), ['json' => $params]);
		}
		catch(\Exception $e)
		{
			$return = null;
		}

		return $return;
	}

	public function createFirewallAccessRule($zoneId, array $configuration, $mode = 'block', $notes = '')
	{
		$params = [
			'mode' => $mode,
			'configuration' => $configuration,
			'notes' => $notes
		];

		try {
			$return = $this->makeRequest('POST', sprintf('zones/%s/firewall/access_rules/rules', $zoneId), ['json' => $params]);
		}
		catch(\Exception $e)
		{
			$return = null;
		}

		return $return;
	}

	public function updateFirewallAccessRule($zoneId, $identifier, $mode, $notes)
	{
		$params = [];

		$params['mode'] = $mode;
		$params['notes'] = $notes;

		try {
			$return = $this->makeRequest('PATCH', sprintf('zones/%s/firewall/access_rules/rules/%s', $zoneId, $identifier), ['json' => $params]);
		}
		catch(\Exception $e)
		{
			$return = null;
		}

		return $return;
	}

	public function deleteFirewallAccessRule($zoneId, $identifier, $cascade = 'none')
	{
		$params = [
			'cascade' => $cascade,
		];
		return $this->makeRequest('DELETE', sprintf('zones/%s/firewall/access_rules/rules/%s', $zoneId, $identifier), ['json' => $params]);
	}

	// This is an Enterprise only API method
	public function traceroute($accountId, $targets, $colos = [], $options = [])
	{
		$params = [
			'targets' => $targets,
		];
		if ($colos)
		{
			$params['colos'] = $colos;
		}
		if ($options)
		{
			$params['options'] = $options;
		}

		return $this->makeRequest('POST', sprintf('accounts/%s/diagnostics/traceroute', $accountId), ['json' => $params]);
	}

	public function createAccessApp($accountId, $domain, $name, array $policies, $sessionDuration = '24h')
	{
		$params = [
			'domain' => $domain,
			'name' => $name,
			'policies' => $policies,
			'session_duration' => $sessionDuration,
			'type' => 'self_hosted'
		];

		try {
			$return = $this->makeRequest('POST', sprintf('accounts/%s/access/apps', $accountId), ['json' => $params]);
		}
		catch(\Exception $e)
		{
			$return = null;
		}
		return $return;
	}

	public function deleteAccessApp($accountId, $identifier)
	{
		return $this->makeRequest('DELETE', sprintf('accounts/%s/access/apps/%s', $accountId, $identifier), ['json' => []]);
	}

	public function getAccessIdentityProviders($accountId)
	{
		$params = [];
		return $this->makeRequest('GET', sprintf('accounts/%s/access/identity_providers', $accountId), ['query' => $params]);
	}

	public function getWorkers($accountId)
	{
		$params = [];
		return $this->makeRequest('GET', sprintf('accounts/%s/workers/scripts', $accountId), ['query' => $params]);
	}

	public function getWorkerSubdomain($accountId)
	{
		$params = [];
		return $this->makeRequest('GET', sprintf('accounts/%s/workers/subdomain', $accountId), ['query' => $params], 10007);
	}

	public function uploadWorker($accountId, $workerId, $code)
	{
		return $this->makeRequest('PUT', sprintf('accounts/%s/workers/scripts/%s', $accountId, $workerId), [
			'multipart' => [
				[
					'name' => 'worker.js',
					'contents' => $code,
					'filename' => 'worker.js',
					'headers'  => ['Content-Type' => 'application/javascript+module']
				],
				[
					'name' => 'metadata',
					'contents' => json_encode([
						'main_module' => 'worker.js'
					])
				]
			],
		]);
	}

	public function getWorkerEnvironmentProductionSubdomain($accountId, $workerId)
	{
		$params = [];
		return $this->makeRequest('GET', sprintf('accounts/%s/workers/services/%s/environments/production/subdomain', $accountId, $workerId), ['query' => $params]);
	}

	public function setWorkerEnvironmentProductionSubdomain($accountId, $workerId, $value)
	{
		$params = [
			'enabled' => $value
		];
		return $this->makeRequest('POST', sprintf('accounts/%s/workers/services/%s/environments/production/subdomain', $accountId, $workerId), ['json' => $params]);
	}

	public function getR2Buckets($accountId)
	{
		$params = ['per_page' => 1000];
		return $this->makeRequest('GET', sprintf('accounts/%s/r2/buckets', $accountId), ['query' => $params]);
	}

	public function createR2Bucket($accountId, $bucketName)
	{
		$params = [
			'name' => $bucketName,
		];
		return $this->makeRequest('POST', sprintf('accounts/%s/r2/buckets', $accountId), ['json' => $params]);
	}

	public function headR2Bucket($bucketName, $returnPromise = false)
	{
		$method = $returnPromise ? 'headAsync' : 'HEAD';

		$params = [];
		return $this->makeRequest($method, '', $params, true, $bucketName);
	}

	public function createR2CustomDomain($accountId, $bucketName, $customDomain, $zoneId, $zoneName)
	{
		$params = [
			'domain' => $customDomain,
			'zoneId' => $zoneId,
			'zoneName' => $zoneName
		];
		return $this->makeRequest('POST', sprintf('accounts/%s/r2/buckets/%s/custom_domains', $accountId, $bucketName), ['json' => $params]);
	}

	public function deleteR2CustomDomain($accountId, $bucketName, $customDomain)
	{
		return $this->makeRequest('DELETE', sprintf('accounts/%s/r2/buckets/%s/custom_domains/%s', $accountId, $bucketName, $customDomain), ['query' => []]);
	}

	public function getR2Usage($accountId, $bucketName, $returnPromise = false)
	{
		$method = $returnPromise ? 'getAsync' : 'GET';
		$params = [];
		return $this->makeRequest($method, sprintf('accounts/%s/r2/buckets/%s/usage', $accountId, $bucketName), $params);
	}

	public function getR2CustomDomains($accountId, $bucketName, $returnPromise = false)
	{
		$method = $returnPromise ? 'getAsync' : 'GET';
		$params = [];
		return $this->makeRequest($method, sprintf('accounts/%s/r2/buckets/%s/custom_domains', $accountId, $bucketName), $params);
	}

	public function getR2Policy($accountId, $bucketName)
	{
		$params = [
		];
		return $this->makeRequest('GET', sprintf('accounts/%s/r2/buckets/%s/policy', $accountId, $bucketName), ['query' => $params]);
	}

	public function setR2Policy($accountId, $bucketName, $customDomain, $access)
	{
		if (!is_array($customDomain))
		{
			$params = [
				'cname' => $customDomain,
				'access' => $access,
			];
		}
		else
		{
			// Bit of a weird situation because the API works by having multiple fo the same `cname` variables, so we need to manually build the query string in this case.
			$params = '';
			foreach($customDomain as $domain)
			{
				$params .= 'cname=' . urlencode($domain) . '&';
			}
			$params .= 'access=' . urlencode($access);
		}

		return $this->makeRequest('PUT', sprintf('accounts/%s/r2/buckets/%s/policy', $accountId, $bucketName), ['query' => $params]);
	}

	public function listR2Bucket($bucketName, $continuationToken = '', $prefix = '', $delimiter = '', $startAfterKey = '')
	{
		$params = [
			'list-type' => 2,
			'encoding-type' => 'url',
			'max-keys' => 1000,
		];
		if ($continuationToken)
		{
			$params['continuation-token'] = $continuationToken;
		}
		if ($prefix)
		{
			$params['prefix'] = $prefix;
		}
		if ($delimiter)
		{
			$params['delimiter'] = $delimiter;
		}
		if ($startAfterKey)
		{
			$params['start-after'] = $startAfterKey;
		}

		return $this->makeRequest('GET', '', ['query' => $params], 0, $bucketName);
	}
	public function getR2Object($bucketName, $filename, $returnPromise = false)
	{
		$method = $returnPromise ? 'getAsync' : 'GET';

		$params = [];
		return $this->makeRequest($method, $filename, $params, 0, $bucketName);
	}

	public function headR2Object($bucketName, $filename, $returnPromise = false)
	{
		$method = $returnPromise ? 'headAsync' : 'HEAD';

		$params = [];
		return $this->makeRequest($method, $filename, $params, true, $bucketName);
	}

	public function createR2Object($bucketName, $filename, $data, $time = 0, $returnPromise = false)
	{
		if (!$time)
		{
			$time = time();
		}

		$params = [
			'body' => $data,
		];

		$method = $returnPromise ? 'putAsync' : 'PUT';
		return $this->makeRequest($method, $filename, $params, 0, $bucketName);

		// In case Cloudflare ever has an API that isn't S3.
		//return $this->makeRequest($method, sprintf('accounts/%s/r2/buckets/%s/objects/%s', $accountId, $bucketName, $filename), ['body' => $data], 0, true);
	}

	public function copyR2Object($bucketName, $filename, $source, $keepMetadata = true)
	{
		$params = [
			'headers' => [
				'x-amz-copy-source' => $source,
				'x-amz-metadata-directive' => ($keepMetadata ? 'COPY' : 'REPLACE')
			],
		];
		return $this->makeRequest('PUT', $filename, $params, 0, $bucketName);
	}
	public function deleteR2Object($bucketName, $filename, $returnPromise = false)
	{
		$method = $returnPromise ? 'deleteAsync' : 'DELETE';
		$params = [];
		return $this->makeRequest($method, $filename, $params, 0, $bucketName);
	}



	/*
	 * The whole signing process for AWS S3 is pretty annoying, hopefully Cloudflare releases their own API at some point.
	 * See: https://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html
	 */
	protected function signS3Request($method, $fullUrl, array &$params)
	{
		$token = $this->option('cloudflareAuth.token');
		if (empty($token))
		{
			$this->printableException($this->phrase('missing_cloudflare_auth_token'));
		}

		$time = time(); // Reading the time here to prevent the off-chance where it changes as we go through signing process

		$params['headers']['X-Amz-Content-Sha256'] = hash('sha256', !empty($params['body']) ? $params['body'] : '');
		$params['headers']['X-Amz-Date'] = gmdate('Ymd\THis\Z', $time);

		// AWS S3 V4
		$xAmzHeaders = [
			'host' => parse_url($fullUrl, PHP_URL_HOST)
		];
		foreach ($params['headers'] as $header => $value)
		{
			$header = trim(strtolower($header));
			if (substr($header, 0, 6) == 'x-amz-')
			{
				$xAmzHeaders[$header] = trim($value);
			}
		}
		ksort($xAmzHeaders);
		$xAmzHeadersString = '';
		foreach ($xAmzHeaders as $header => $value)
		{
			$xAmzHeadersString .= $header . ':' . $value . "\n";
		}

		if (!empty($params['query']))
		{
			ksort($params['query']);
		}

		$canonicalRequest = strtoupper(str_replace('Async', '', $method)) . "\n" .
			str_replace('%2F', '/', rawurlencode(parse_url($fullUrl, PHP_URL_PATH))) . "\n" .
			(empty($params['query']) ? '' : http_build_query($params['query'])) . "\n" .
			$xAmzHeadersString . "\n" .
			implode(';', array_keys($xAmzHeaders)) . "\n" .
			$params['headers']['X-Amz-Content-Sha256'];

		$stringToSign = "AWS4-HMAC-SHA256\n".
			$params['headers']['X-Amz-Date'] . "\n" .
			gmdate('Ymd', $time) . "/auto/s3/aws4_request\n" .
			hash('sha256', $canonicalRequest);

		$params['headers']['Authorization'] = 'AWS4-HMAC-SHA256 Credential=' . $this->option('cfTokenId') . '/' . gmdate('Ymd', $time) . '/auto/s3/aws4_request, ' .
			'SignedHeaders=' . implode(';', array_keys($xAmzHeaders)) . ', ' .
			'Signature=' . hash_hmac('sha256', $stringToSign, $this->getSigningKey($token, $time));
	}

	protected function getSigningKey($token, $time, $region = 'auto', $service = 's3')
	{
		$dateKey = hash_hmac('sha256', gmdate('Ymd', $time), 'AWS4' . hash('sha256', $token), true);
		$dateRegionKey = hash_hmac('sha256', $region, $dateKey, true);
		$dateRegionServiceKey = hash_hmac('sha256', $service, $dateRegionKey, true);
		return hash_hmac('sha256', 'aws4_request', $dateRegionServiceKey, true);
	}



	/*
	 * A work in progress...
	 *
	 * See some info here in presign() method:  https://github.com/aws/aws-sdk-php/blob/3ec607c6280a1991bcb76cfac59c089da962c3ee/src/Signature/SignatureV4.php
	 *
	 * https://docs.aws.amazon.com/AmazonCloudFront/latest/DeveloperGuide/CreateURL_PHP.html
	 */
	function getPresignedUrl($method, $bucketName, $objectPath, $secondsUntilExpiration = 60, array $extraQuery = [], array $extraHeader = [], $limitToIp = null)
	{
		$token = $this->option('cloudflareAuth.token');
		if (empty($token))
		{
			$this->printableException($this->phrase('missing_cloudflare_auth_token'));
		}

		$time = time();
		/*
				$policyCondition = [];
				if ($limitToIp)
				{
					$policyCondition['IpAddress'] = [
						'AWS:SourceIp' => $limitToIp . '/32'
					];
				}
				if ($secondsUntilExpiration)
				{
					$policyCondition['DateLessThan'] = [
						'AWS:EpochTime' => time() + $secondsUntilExpiration
					];
				}


				$policy = [
					'Statement' => [
						[
							'Resource' => $objectPath,
						]
					]
				];

				if ($policyCondition)
				{
					$policy['Statement'][0]['Condition'] = $policyCondition;
				}

				echo "generated policy:\r\n" . json_encode($policy);

				$encodedPolicy = str_replace(
					['+', '=', '/'],
					['-', '_', '~'],
					base64_encode(json_encode($policy))
				);
		*/


		$query = [
			'X-Amz-Algorithm' => 'AWS4-HMAC-SHA256',
			'X-Amz-Credential' => $this->option('cfTokenId') . '/' . gmdate('Ymd', $time) . '/auto/s3/aws4_request',
			'X-Amz-Date' => gmdate('Ymd\THis\Z', $time),
			'X-Amz-Expires' => $secondsUntilExpiration,
			//	'X-Amz-SignedHeaders' => implode(';', $headers),
		];

		$baseUrl = $this->r2BaseUrlFull($bucketName);

		// AWS S3 V4
		$xAmzHeaders = [
			'host' => parse_url($baseUrl, PHP_URL_HOST)
		];

		if (count($extraQuery))
		{
			$query = array_merge($query, $extraQuery);
		}

		if (count($extraHeader))
		{
			$xAmzHeaders = array_merge($xAmzHeaders, $extraHeader);
		}

		// Even though S3 documentation says it should be all x-amz-* headers, it seems it's only Host (maybe because presigned URLs are in query and not actually in headers)?
		/*
				foreach ($query as $header => $value)
				{
					$header = trim(strtolower($header));
					if (substr($header, 0, 6) == 'x-amz-')
					{
						$xAmzHeaders[$header] = trim($value);
					}
				}
				ksort($xAmzHeaders);
		*/

		$xAmzHeadersString = '';
		foreach ($xAmzHeaders as $header => $value)
		{
			$xAmzHeadersString .= $header . ':' . $value . "\n";
		}

		$query['X-Amz-SignedHeaders'] = implode(';', array_keys($xAmzHeaders));

		if (!empty($query))
		{
			ksort($query);
		}

		$canonicalRequest = strtoupper(str_replace('Async', '', $method)) . "\n" .
			'/' . $bucketName . '/' . str_replace('%2F', '/', rawurlencode(parse_url($objectPath, PHP_URL_PATH))) . "\n" .
			(empty($query) ? '' : http_build_query($query, '', '&', PHP_QUERY_RFC3986)) . "\n" .
			$xAmzHeadersString . "\n" .
			implode(';', array_keys($xAmzHeaders)) . "\n" .
			'UNSIGNED-PAYLOAD';

		$stringToSign = "AWS4-HMAC-SHA256\n".
			$query['X-Amz-Date'] . "\n" .
			gmdate('Ymd', $time) . "/auto/s3/aws4_request\n" .
			hash('sha256', $canonicalRequest);

		$query['X-Amz-Signature'] = hash_hmac('sha256', $stringToSign, $this->getSigningKey($token, $time));

		return $baseUrl . $objectPath . '?' . http_build_query($query, '', '&', PHP_QUERY_RFC3986);
	}

}