<?php

namespace DigitalPoint\Cloudflare\Cli\Command;

use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Helper\ProgressBar;
use Symfony\Component\Console\Helper\QuestionHelper;
use Symfony\Component\Console\Helper\Table;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Question\ChoiceQuestion;
use Symfony\Component\Console\Question\ConfirmationQuestion;
use Symfony\Component\Process\Process;
use Symfony\Component\Process\PhpExecutableFinder;
class MigrateData extends Command
{
	protected $startTime = null;
	protected $totalSize = null;
	protected $finishedSize = null;
	protected $progressBar = null;

	protected function configure()
	{
		$this
			->setName('dp:migrate-data')
			->setDescription('Migrate data between two different abstracted filesystems.')
			->addOption(
				'new-to-old',
				null,
				InputOption::VALUE_NONE,
				'If set, data will move in the opposite direction (from the "new" adapter to the "old" adapter).'
			)
			->addOption(
				'processes',
				null,
				InputOption::VALUE_REQUIRED,
				'If set to a value between 2 and the number of CPU cores your serer has, files will be moved concurrently with this many processes.'
			)
			->addOption(
				'start-at-path',
				null,
				InputOption::VALUE_REQUIRED,
				'If set, files before this path will be skipped.  Example: attachments/0/123-0123456789abcdef0123456789abcdef.data'
			)
			->addOption(
				'location',
				null,
				InputOption::VALUE_REQUIRED,
				'You can skip the location prompt by specifying it here.  Example: data'
			)
			->addOption(
				'path',
				null,
				InputOption::VALUE_REQUIRED,
				'If set, only this file will be migrated.  Example: attachments/0/123-0123456789abcdef0123456789abcdef.data'
			);
	}

	protected function execute(InputInterface $input, OutputInterface $output)
	{
		/** @var \DigitalPoint\Cloudflare\XF\FsMounts $mountsClass */
		$mountsClass = \XF::extendClass('XF\FsMounts');

		/** @var \DigitalPoint\Cloudflare\League\Flysystem\MountManager $fromFs */

		if ($input->getOption('new-to-old'))
		{
			$newToOld = true;
			$fromFs = \XF::fs();
			$mountsClass::$allowAdaptersNotInConfig = false;
			$toFs = $mountsClass::loadDefaultMounts(\XF::config());
			$mountsClass::$allowAdaptersNotInConfig = true;
		}
		else
		{
			$newToOld = false;
			$mountsClass::$allowAdaptersNotInConfig = false;
			$fromFs = $mountsClass::loadDefaultMounts(\XF::config());
			$mountsClass::$allowAdaptersNotInConfig = true;
			$toFs = \XF::fs();
		}

		/** @var \DigitalPoint\Cloudflare\Repository\Cli $cliRepo */
		$cliRepo = \XF::repository('DigitalPoint\Cloudflare:Cli');

		$fsMapKeys = [];
		$fsMap = $cliRepo->getFsMap($fromFs, $toFs, $fsMapKeys);

		$location = $input->getOption('location');
		$path = $input->getOption('path');

		// This is normally a child process running in parallel
		if ($location && $path)
		{
			$config = new \League\Flysystem\Config();

			$data = $fsMap[$location]['from']->read($path);
			$fsMap[$location]['to']->write($path, $data['contents'], $config);

			return 1;
		}

		$output->writeln('');
		$output->writeln('<comment>' . \XF::phrase('cf_migrate_data_note') . '</comment>');

		// Show new and old adapters
		$output->writeln('');
		$table = new Table($output);
		$table->setHeaders([
			[\XF::phrase('location'), \XF::phrase('from_adapter'), \XF::phrase('to_adapter')]
		]);
		foreach ($fsMap as $name => $info)
		{
			$table->addRow([$name, get_class($info['from']), get_class($info['to'])]);
		}
		$table->render();
		$output->writeln('');

		/** @var QuestionHelper $helper */
		$helper = $this->getHelper('question');

		if (!$location)
		{
			$question = new ChoiceQuestion(
				'<question>' . \XF::phrase('which_location_would_you_like_to_migrate') . '</question>',
				$fsMapKeys
			);

			$location = $helper->ask($input, $output, $question);
			$output->writeln('');
		}

		if (get_class($fsMap[$location]['from']) == get_class($fsMap[$location]['to']))
		{
			$output->writeln('<error>' . \XF::phrase('filesystem_same_adapters_x', ['adapter' => get_class($fsMap[$location]['from'])]) . '</error>');
			exit;
		}

		$files = $this->getFiles($fsMap[$location]['from'], $location);

		$question = new ConfirmationQuestion(
			\XF::phrase('location_x', ['location' => $location]) . '
' . \XF::phrase('from_x', ['from' => get_class($fsMap[$location]['from'])]) . ' 
' . \XF::phrase('to_x', ['to' => get_class($fsMap[$location]['to'])]) . '
' . \XF::phrase('memory_used_x_allowed_y', ['used' => \XF::language()->fileSizeFormat(memory_get_usage()), 'allowed' => ini_get('memory_limit') >= 0 ? ini_get('memory_limit') : '∞' ]) . '
<comment>' . \XF::phrase('cf_migrate_data_delete_from_note') . '</comment>
<question>' . \XF::phrase('are_you_certain_you_want_to_migrate_x_y', ['count' => \XF::language()->numberFormat(count($files)), 'size' => \XF::language()->numberFormat($this->totalSize)]) . '</question>'
		);

		if ($helper->ask($input, $output, $question))
		{
			$this->startTime = microtime(true);

			$this->progressBar = new ProgressBar($output, count($files));
			$this->progressBar->setFormat(' %current%/%max% [%bar%] %percent:3s%%  %elapsed:6s%  %message%/s');

			$startAtPath = $input->getOption('start-at-path');
			$processes = intval($input->getOption('processes'));

			if ($processes < 1)
			{
				$processes = 1;
			}
			else
			{
				if (strtoupper(substr(PHP_OS, 0, 3)) === 'WIN')
				{
					$cores = intval(shell_exec('echo %NUMBER_OF_PROCESSORS%'));
				}
				else
				{
					$cores = intval(shell_exec('nproc'));
				}

				if ($cores > 0 && $cores < $processes)
				{
					$processes = $cores;
				}
			}

			$output->writeln('');

			if ($processes > 1)
			{
				$threads = [];

				$processFinder = new PhpExecutableFinder();
				$parallelProcessCommand = $processFinder->find(false) . ' ' . $_SERVER['argv'][0] . ' '. $_SERVER['argv'][1] . ($newToOld ? ' --new-to-old' : '') . ' --location=' . $location . ' --path=';

				foreach($files as $file)
				{
					if ($startAtPath)
					{
						if ($file['path'] == $startAtPath)
						{
							$startAtPath = '';
						}
						else
						{
							$this->advanceProgressBar();
							continue;
						}
					}

					if (count($threads) >= $processes)
					{
						$threads = $this->waitForProcesses($threads);
					}

					$process = new Process($parallelProcessCommand . $file['path']);
					$process->start();
					$threads[] = $process;

					$this->finishedSize += $file['size'];
				}

				$this->waitForProcesses($threads);
			}
			else
			{
				$adapterFrom = $fsMap[$location]['from'];
				$adapterTo = $fsMap[$location]['to'];
				$config = new \League\Flysystem\Config();

				foreach($files as $file)
				{
					if ($startAtPath)
					{
						if ($file['path'] == $startAtPath)
						{
							$startAtPath = '';
						}
						else
						{
							$this->advanceProgressBar();
							continue;
						}
					}

					$data = $adapterFrom->read($file['path']);
					$adapterTo->write($file['path'], $data['contents'], $config);
					$this->finishedSize += $file['size'];
					$this->advanceProgressBar();
				}
			}

			$this->setMessageProgressBar();
			$this->progressBar->finish();
		}

		$output->writeln('');
		return 1;
	}

	protected function getFiles(\League\Flysystem\Adapter\AbstractAdapter $adapter, $location)
	{
		$path = '';

		if (substr_count($location, '/') != 0)
		{
			$split = explode('/', $location, 2);
			$path = $split[1];
		}

		$files = $adapter->listContents($path, true);

		$this->totalSize = 0;

		$files = array_filter($files, function($file) {
			if ($file['type'] == 'file' && substr($file['path'], -10) != 'index.html')
			{
				$this->totalSize += $file['size'];
				return true;
			}
			return false;
		});

		return array_values($files);
	}

	protected function waitForProcesses(array $threads)
	{
		foreach ($threads as $thread)
		{
			$thread->wait();
			$this->advanceProgressBar();
		}
		return [];
	}

	protected function setMessageProgressBar()
	{
		$this->progressBar->setMessage(\XF::language()->fileSizeFormat($this->finishedSize / (microtime(true) - $this->startTime)));
	}

	protected function advanceProgressBar()
	{
		$this->setMessageProgressBar();
		$this->progressBar->advance();
	}
}