<?php

namespace DigitalPoint\Cloudflare\XF;

use League\Flysystem\AdapterInterface;
use League\Flysystem\EventableFilesystem\EventableFilesystem;

class FsMounts extends XFCP_FsMounts
{
	/**
	 * Not the ideal way to do this, but need to keep loadDefaultMounts() method compatible with other addons.  Only used in the dp:migrate-data CLI tool.
	 */
	public static $allowAdaptersNotInConfig = true;

	/**
	 * This is based on the XenForo standard method, with some extra stuff:
	 * - Ability to add adapters based on admin options with static::getDpFsAdapters() method (allows setup of adapters via UI)
	 * - Allows disable_asserts to be set on a per-adapter basis
	 * - Automatically merging "sub-adapters" from config.php
	 * - Returning an extended version of \League\Flysystem\MountManager (sub-adapter support)
	 *
	 * @param array $config
	 *
	 * @return \DigitalPoint\Cloudflare\League\Flysystem\MountManager
	 */
	public static function loadDefaultMounts(array $config)
	{
		$adapterOverrides = static::getDpFsAdapters($config, static::$allowAdaptersNotInConfig);

		if (isset($adapterOverrides['data']))
		{
			$dataAdapter = call_user_func($adapterOverrides['data']);
		}
		else
		{
			$dataAdapter = static::getLocalAdapter($config['externalDataPath']);
		}

		$data = new EventableFilesystem($dataAdapter, [
			'visibility' => AdapterInterface::VISIBILITY_PUBLIC,
			'disable_asserts' => (property_exists($dataAdapter, 'disableAsserts') ? $dataAdapter->disableAsserts : false)
		]);
		static::addDefaultWriteListeners('data', $data);

		if (isset($adapterOverrides['internal-data']))
		{
			$internalDataAdapter = call_user_func($adapterOverrides['internal-data']);
		}
		else
		{
			$internalDataAdapter = static::getLocalAdapter($config['internalDataPath']);
		}

		$internalData = new EventableFilesystem($internalDataAdapter, [
			'visibility' => AdapterInterface::VISIBILITY_PUBLIC,
			'disable_asserts' => (property_exists($internalDataAdapter, 'disableAsserts') ? $internalDataAdapter->disableAsserts : false)
		]);
		static::addDefaultWriteListeners('internal-data', $internalData);

		if (!empty($config['localDataPath'])) // Would only be there in XenForo 2.3+
		{
			if (isset($adapterOverrides['local-data']))
			{
				$localDataAdapter = call_user_func($adapterOverrides['local-data']);
			}
			else
			{
				$localDataAdapter = static::getLocalAdapter($config['localDataPath']);
			}

			$localData = new EventableFilesystem($localDataAdapter, [
				'visibility' => AdapterInterface::VISIBILITY_PUBLIC,
				'disable_asserts' => (property_exists($localDataAdapter, 'disableAsserts') ? $localDataAdapter->disableAsserts : false)
			]);

			static::addDefaultWriteListeners('local-data', $data);
		}

		// if you override this, codeCachePath must still contain the code as we will use it to include files
		if (isset($adapterOverrides['code-cache']))
		{
			$codeCacheAdapter = call_user_func($adapterOverrides['code-cache']);
		}
		else
		{
			$codeCacheAdapter = static::getLocalAdapter(
				sprintf($config['codeCachePath'], $config['internalDataPath'])
			);
		}

		$codeCache = new EventableFilesystem($codeCacheAdapter, [
			'visibility' => AdapterInterface::VISIBILITY_PUBLIC,
			'disable_asserts' => (property_exists($codeCacheAdapter, 'disableAsserts') ? $codeCacheAdapter->disableAsserts : false)
		]);
		static::addDefaultWriteListeners('code-cache', $codeCache);

		$subAdapters = [];

		if ($adapterOverrides)
		{
			foreach($adapterOverrides as $prefix => $getAdapterMethod)
			{
				// Ignore the 3 standard prefixes that XenForo uses since they have already been handled
				if (in_array($prefix, ['data', 'internal-data', 'code-cache', 'local-data']))
				{
					continue;
				}

				$adapter = call_user_func($getAdapterMethod);
				$subAdapters[$prefix] = new EventableFilesystem($adapter, [
					'visibility' => AdapterInterface::VISIBILITY_PUBLIC,
					'disable_asserts' => (property_exists($adapter, 'disableAsserts') ? $adapter->disableAsserts : false)
				]);
				static::addDefaultWriteListeners($prefix, $subAdapters[$prefix]);
			}
		}

		/** @var \DigitalPoint\FileSystem\League\Flysystem\MountManager $mountManagerClass */
		$mountManagerClass = \XF::extendClass('DigitalPoint\Cloudflare\League\Flysystem\MountManager');

		$adapters = array_merge(
			[
				'data' => $data,
				'internal-data' => $internalData,
	//			'local-data' => $localData,
				'code-cache' => $codeCache
			],
			$subAdapters
		);

		if (!empty($localData))
		{
			$adapters['local-data'] = $localData;
		}

		return new $mountManagerClass($adapters);
	}

	/**
	 * Allows filesystem adapters to be injected without explicitly adding them to config.php
	 *
	 * @param array $config
	 *
	 * @return array
	 */
	protected static function getDpFsAdapters(array $config, $allowAdaptersNotInConfig = true)
	{
		if ($allowAdaptersNotInConfig)
		{
			foreach (\XF::options()->cfR2Bucket as $path => $bucket)
			{
				if (!empty($bucket))
				{
					$config['fsAdapters'][$path] = function () use ($path)
					{
						return \DigitalPoint\Cloudflare\League\Flysystem\Adapter\R2::getAdapter($path);
					};
				}
			}
		}

		return $config['fsAdapters'];
	}
}

// ******************** FOR IDE AUTO COMPLETE ********************
if (false)
{
	class XFCP_FsMounts extends \XF\FsMounts {}
}